package AA::Vertex;

our @ISA = ('AA');

sub new($$) {
	my ($class, $id) = @_;
	my $v = AA::new($class, $id);
	$v->DefSet('x', 100);
	$v->DefSet('y', 100);
	$v->DefSet('color', [ 0.7, 0, 0.7 ]);
	$v->DefSet('gradient', 0.5);
	$v->DefSet('r', 30);
	$v->DefSet('rim', 2);
	$v->DefSet('rim-color', [ 0.1, 0, 0.1 ]);
	$v->Define('label');
	$v->DefSet('label-color', [ 1, 1, 1 ]);
	$v->DefSet('label-size', 1.5);
	$v->DefSet('d', 20);
	return $v;
}

sub Draw($$) {
	my ($v, $cairo) = @_;
	my ($x, $y, $r) = ($v->Get('x'), $v->Get('y'), $v->Get('r'));
	my $grad = $v->Get('gradient');
	my @col = @{$v->Get('color')};

	$cairo->new_path;

	if ($grad) {
		my $lg = Cairo::RadialGradient->create($x, $y, 0, $x, $y, $r);
		$lg->add_color_stop_rgb(0, @col);
		$col[0] *= $grad;
		$col[1] *= $grad;
		$col[2] *= $grad;
		$lg->add_color_stop_rgb(1, @col);
		$cairo->set_source($lg);
	} else {
		$cairo->set_source_rgb(@col);
	}
	$cairo->arc($x, $y, $r, 0, $AA::TWOPI);
	$cairo->fill_preserve;

	$cairo->set_source_rgb(@{$v->Get('rim-color')});
	$cairo->set_line_width($v->Get('rim'));
	$cairo->stroke;

	if ($v->Exists('label')) {
		my $label = $v->Get('label');
		$cairo->set_source_rgb(@{$v->Get('label-color')});
		$cairo->set_font_size($r * $v->Get('label-size'));
		my $xt = $cairo->text_extents($label);
		$cairo->move_to($x - $xt->{'width'}/2 - $xt->{'x_bearing'},
				$y - $xt->{'height'}/2 - $xt->{'y_bearing'});
		$cairo->show_text($label);
	}
}

package AA::Edge;

our @ISA = ('AA');

sub new($$$$) {
	my ($class, $id, $v, $w) = @_;
	my $e = AA::new($class, $id);
	$e->DefSet('color', [1, 1, 1]);
	$e->DefSet('width', 3);
	$e->DefSet('arrow', 0);
	$e->DefSet('arrow-size', 20);	# distance between arrow tip and back
	$e->DefSet('arrow-span', 0.5);	# fraction of arrow-size
	$e->DefSet('overshoot', 3);
	$e->DefBind('dir', sub {
		my $dx = $w->Get('x') - $v->Get('x');
		my $dy = $w->Get('y') - $v->Get('y');
		my $d = sqrt($dx*$dx + $dy*$dy);
		if ($d) {
			return [$dx/$d, $dy/$d];
		} else {
			return [0, 0];
		}
	});
	$e->DefBind('ends', sub {
		my $over = $e->Get('overshoot');
		my $x1 = $v->Get('x');
		my $y1 = $v->Get('y');
		my $r1 = $v->Get('r') - $over;
		my $x2 = $w->Get('x');
		my $y2 = $w->Get('y');
		my $r2 = $w->Get('r') - $over;
		my ($dx,$dy) = @{$e->Get('dir')};
		return [ $x1+$r1*$dx, $y1+$r1*$dy, $x2-$r2*$dx, $y2-$r2*$dy ];
	});
	$e->DefSet('d', 10);
	return $e;
}

sub Draw($$) {
	my ($e, $cairo) = @_;
	my ($x1, $y1, $x2, $y2) = @{$e->Get('ends')};
	$cairo->set_source_rgb(@{$e->Get('color')});
	$cairo->set_line_width($e->Get('width'));
	if (my $arrow = $e->Get('arrow')) {
		my $size = $e->Get('arrow-size');
		my $span = $size * $e->Get('arrow-span');
		my ($dx,$dy) = @{$e->Get('dir')};
		# line
		$cairo->move_to($x1, $y1);
		$cairo->line_to($x2 - $dx*$size*0.9, $y2 - $dy*$size*0.9);
		$cairo->stroke;
		# arrow
		my ($ax,$ay) = (-$dy,$dx);
		my ($bx,$by) = ($x2 - $dx*$size, $y2 - $dy*$size);
		$cairo->move_to($x2, $y2);
		$cairo->line_to($bx + $ax*$span, $by + $ay*$span);
		$cairo->line_to($bx - $ax*$span, $by - $ay*$span);
		$cairo->close_path;
		$cairo->fill;
	} else {
		$cairo->move_to($x1, $y1);
		$cairo->line_to($x2, $y2);
		$cairo->stroke;
	}
}

1;
